@@ -1,9 +1,9 @@
 #include "Rcpp.h"
 #include "BiocNeighbors.h"
 
-// Turn off manual vectorization always, to avoid small inconsistencies in
-// distance calculations across otherwise-compliant machines. 
-#define NO_MANUAL_VECTORIZATION 1
+#include <string>
+#include <memory>
+#include <stdexcept>
 
 // Copied from RcppAnnoy's RcppAnnoy.h:
 #if defined(__MINGW32__)
