@@ -1,16 +1,91 @@
 #include "Rcpp.h"
 #include "BiocNeighbors.h"
+#include "Rtatami.h"
+
 #include "knncolle/knncolle.hpp"
+#include "knncolle_tatami/knncolle_tatami.hpp"
 
 #include <stdexcept>
 #include <string>
 #include <algorithm>
 #include <vector>
+#include <optional>
+#include <cmath>
+
+void check_nonfinite(const double* ptr, std::size_t n) {
+    std::size_t count = 0; 
+    for (std::size_t i = 0; i < n; ++i) {
+        count += !std::isfinite(ptr[i]);
+    }
+    if (count) { // checking once at the end, to allow vectorization of the loop. 
+        throw std::runtime_error("non-finite values are not supported");
+    }
+}
+
+class NonfiniteMatrixExtractor final : public knncolle::MatrixExtractor<double> {
+public:
+    NonfiniteMatrixExtractor(std::unique_ptr<knncolle::MatrixExtractor<double> > ptr, std::size_t ndim) :
+        my_ptr(std::move(ptr)),
+        my_ndim(ndim)
+    {}
+
+private:
+    std::unique_ptr<knncolle::MatrixExtractor<double> > my_ptr;
+    std::size_t my_ndim;
+
+public:
+    const double* next() {
+        const auto output = my_ptr->next();
+        check_nonfinite(output, my_ndim);
+        return output;
+    }
+};
+
+class NonfiniteChecker final : public knncolle::Matrix<int, double> {
+public:
+    NonfiniteChecker(std::unique_ptr<knncolle::Matrix<int, double> > ptr) : my_ptr(std::move(ptr)) {}
+
+private:
+    std::unique_ptr<knncolle::Matrix<int, double> > my_ptr;
+
+public:
+    int num_observations() const {
+        return my_ptr->num_observations();
+    }
+
+    std::size_t num_dimensions() const {
+        return my_ptr->num_dimensions();
+    }
+
+    std::unique_ptr<knncolle::MatrixExtractor<double> > new_extractor() const {
+        return std::make_unique<NonfiniteMatrixExtractor>(my_ptr->new_extractor(), my_ptr->num_dimensions());
+    }
+};
 
 //[[Rcpp::export(rng=false)]]
-SEXP generic_build(SEXP builder, Rcpp::NumericMatrix data) {
+SEXP generic_build(SEXP builder, SEXP data, bool fail_nonfinite) {
+    std::optional<Rcpp::NumericMatrix> mat;
+    std::unique_ptr<knncolle::Matrix<int, double> > matptr;
+
+    switch (TYPEOF(data)) {
+        case REALSXP: case INTSXP: case LGLSXP: // allow any coercible-to-double type, for back-compatibility.
+            mat = Rcpp::NumericMatrix(data);
+            matptr.reset(new knncolle::SimpleMatrix<int, double>(mat->rows(), mat->cols(), mat->begin()));
+            break;
+        case EXTPTRSXP:
+            matptr.reset(new knncolle_tatami::Matrix<int, double, double, int, const tatami::NumericMatrix*>(Rtatami::BoundNumericPointer(data)->get(), false));
+            break;
+        default:
+            throw std::runtime_error("unknown type for 'data'");
+    }
+
+    if (fail_nonfinite) {
+        auto newmat = std::make_unique<NonfiniteChecker>(std::move(matptr));
+        matptr = std::move(newmat);
+    }
+
     auto out = BiocNeighbors::BuilderPointer(builder);
-    return BiocNeighbors::PrebuiltPointer(out->build_raw(knncolle::SimpleMatrix<int, double>(data.rows(), data.cols(), data.begin())));
+    return BiocNeighbors::PrebuiltPointer(out->build_raw(*matptr));
 }
 
 //[[Rcpp::export(rng=false)]]
@@ -192,17 +267,67 @@ SEXP generic_find_knn(
     }
 } 
 
+struct QueryInfo {
+    QueryInfo(SEXP data) {
+        switch (TYPEOF(data)) {
+            case REALSXP: case INTSXP: case LGLSXP: // allow any coercible-to-double type, for back-compatibility.
+                mat = Rcpp::NumericMatrix(data);
+                data_ptr = mat->begin();
+                nobs = mat->ncol();
+                ndim = mat->nrow();
+                break;
+            case EXTPTRSXP:
+                tatami_ptr = Rtatami::BoundNumericPointer(data)->get();
+                nobs = tatami_ptr->ncol();
+                ndim = tatami_ptr->nrow();
+                break;
+            default:
+                throw std::runtime_error("unknown type for 'data'");
+        }
+    }
+
+    int nobs = 0;
+    std::size_t ndim = 0;
+    std::optional<Rcpp::NumericMatrix> mat;
+    const double* data_ptr = NULL;
+    const tatami::NumericMatrix* tatami_ptr = NULL;
+};
+
+class QueryWorkspace {
+public:
+    QueryWorkspace(const QueryInfo& qinfo, int start, int length) : my_qinfo(qinfo) {
+        if (!qinfo.mat.has_value()) {
+            my_buffer.resize(qinfo.ndim);
+            my_ext = tatami::consecutive_extractor<false, double, int>(*(qinfo.tatami_ptr), false, start, length);
+        }
+    }
+
+    const double* get(int i) {
+        if (my_qinfo.mat.has_value()) {
+            return my_qinfo.data_ptr + static_cast<std::size_t>(i) * my_qinfo.ndim;
+        } else {
+            return my_ext->fetch(my_buffer.data());
+        }
+    }
+
+private:
+    const QueryInfo& my_qinfo;
+    std::vector<double> my_buffer;
+    std::unique_ptr<tatami::OracularDenseExtractor<double, int> > my_ext;
+};
+
 //[[Rcpp::export(rng=false)]]
 SEXP generic_query_knn(
     SEXP prebuilt_ptr,
-    Rcpp::NumericMatrix query,
+    SEXP query,
     Rcpp::IntegerVector num_neighbors,
     bool force_variable_neighbors,
     int num_threads,
     bool last_distance_only,
     bool report_index,
-    bool report_distance)
-{
+    bool report_distance,
+    bool fail_nonfinite
+) {
     BiocNeighbors::PrebuiltPointer cast(prebuilt_ptr); 
     if (!R_ExternalPtrAddr(SEXP(cast))) {
         throw std::runtime_error("null pointer to a prebuilt index");
@@ -211,9 +336,9 @@ SEXP generic_query_knn(
     int nobs = prebuilt.num_observations();
     size_t ndim = prebuilt.num_dimensions();
 
-    int nquery = query.ncol();
-    const double* query_ptr = query.begin();
-    if (static_cast<size_t>(query.nrow()) != ndim) {
+    QueryInfo qinfo(query);
+    const auto nquery = qinfo.nobs;
+    if (qinfo.ndim != ndim) {
         throw std::runtime_error("mismatch in dimensionality between index and 'query'");
     }
 
@@ -275,10 +400,15 @@ SEXP generic_query_knn(
         std::vector<int> tmp_i;
         std::vector<double> tmp_d;
 
-        size_t query_offset = static_cast<size_t>(start) * ndim; // using size_t to avoid overflow.
-        for (int o = start, end = start + length; o < end; ++o, query_offset += ndim) {
+        QueryWorkspace qwork(qinfo, start, length);
+        for (int o = start, end = start + length; o < end; ++o) {
+            const auto qptr = qwork.get(o);
+            if (fail_nonfinite) {
+                check_nonfinite(qptr, ndim);
+            }
+
             searcher->search(
-                query_ptr + query_offset,
+                qptr,
                 (is_k_variable ? variable_k[o] : const_k),
                 (report_index ? &tmp_i : NULL),
                 (report_distance ? &tmp_d : NULL)
@@ -410,7 +540,7 @@ SEXP generic_find_all(SEXP prebuilt_ptr, Rcpp::Nullable<Rcpp::IntegerVector> cho
 } 
 
 //[[Rcpp::export(rng=false)]]
-SEXP generic_query_all(SEXP prebuilt_ptr, Rcpp::NumericMatrix query, Rcpp::NumericVector thresholds, int num_threads, bool report_index, bool report_distance) {
+SEXP generic_query_all(SEXP prebuilt_ptr, SEXP query, Rcpp::NumericVector thresholds, int num_threads, bool report_index, bool report_distance, bool fail_nonfinite) {
     BiocNeighbors::PrebuiltPointer cast(prebuilt_ptr); 
     if (!R_ExternalPtrAddr(SEXP(cast))) {
         throw std::runtime_error("null pointer to a prebuilt index");
@@ -418,9 +548,9 @@ SEXP generic_query_all(SEXP prebuilt_ptr, Rcpp::NumericMatrix query, Rcpp::Numer
     const auto& prebuilt = *cast;
     size_t ndim = prebuilt.num_dimensions();
 
-    int nquery = query.ncol();
-    const double* query_ptr = query.begin();
-    if (static_cast<size_t>(query.nrow()) != ndim) {
+    QueryInfo qinfo(query);
+    const auto nquery = qinfo.nobs;
+    if (qinfo.ndim != ndim) {
         throw std::runtime_error("mismatch in dimensionality between index and 'query'");
     }
 
@@ -449,11 +579,15 @@ SEXP generic_query_all(SEXP prebuilt_ptr, Rcpp::NumericMatrix query, Rcpp::Numer
             return;
         }
 
-        size_t query_offset = static_cast<size_t>(start) * ndim; // using size_t to avoid overflow.
-        for (int o = start, end = start + length; o < end; ++o, query_offset += ndim) {
-            auto current_ptr = query_ptr + query_offset;
+        QueryWorkspace qwork(qinfo, start, length);
+        for (int o = start, end = start + length; o < end; ++o) {
+            const auto qptr = qwork.get(o);
+            if (fail_nonfinite) {
+                check_nonfinite(qptr, ndim);
+            }
+
             auto count = searcher->search_all(
-                current_ptr,
+                qptr,
                 threshold_ptr[multiple_thresholds ? o : 0],
                 (report_index ? &out_i[o] : NULL),
                 (report_distance ? &out_d[o] : NULL)
