@@ -17,7 +17,11 @@ test_that("Cursory checks for KmknnParam", {
     expect_identical(ncol(out$distance), 8L)
     expect_identical(ncol(out$index), 8L)
 
-    d <- median(out$distance[,ncol(out$distance)])
+    # Slightly increase distance to avoid a direct comparison to the median.
+    # Otherwise, we'd have one distance that is equal to the median,
+    # and that equality might be broken by floating-point imprecision.
+    d <- median(out$distance[,ncol(out$distance)]) * 1.000001
+
     allout <- findNeighbors(Y, threshold=d, BNPARAM=p)
     expect_identical(length(allout$distance), nrow(Y))
     expect_identical(length(allout$index), nrow(Y))
@@ -35,7 +39,7 @@ test_that("KmknnParam queries behave with cosine distance", {
     ref <- queryKNN(Y1, Z1, k=8, BNPARAM=p)
     expect_equal(out, ref)
 
-    d <- median(out$distance[,ncol(out$distance)])
+    d <- median(out$distance[,ncol(out$distance)]) * 1.000001
     allout <- queryNeighbors(Y, Z, threshold=d, BNPARAM=p)
     Y1 <- Y/sqrt(rowSums(Y^2))
     Z1 <- Z/sqrt(rowSums(Z^2))
